unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtDlgs, StdCtrls, ExtCtrls, pngimage, ComCtrls;

type
  PWordArray = ^TWordArray;
  TWordArray = array[0..0] of Word;

  TColorRec = packed record
                Value: Word;
                NewValue: Word;
                Count: Integer;
              end;

  TColorTable = array of TColorRec;

  TForm1 = class(TForm)
    Button1: TButton;
    Button2: TButton;
    Panel1: TPanel;
    Image1: TImage;
    Panel2: TPanel;
    Image2: TImage;
    Button3: TButton;
    Label1: TLabel;
    Button4: TButton;
    Label2: TLabel;
    SD: TSaveDialog;
    M: TMemo;
    OPD: TOpenPictureDialog;
    CB: TCheckBox;
    Edit1: TEdit;
    UpDown1: TUpDown;
    Label3: TLabel;
    procedure Button1Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure CBColorClick(Sender: TObject);
  private
    FStream: TMemoryStream;
    procedure Convert;
    procedure Compress;
    function ColorTable: TColorTable;
  public
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}
uses JPeg;

function TForm1.ColorTable: TColorTable;

  function Compare(const A,C: TColorRec): Integer;
  begin
    if A.Value < C.Value then Result := -1 else
      if A.Value > C.Value then Result := +1
        else Result := 0;
  end;

  procedure Sort(Colors: TColorTable; L,R: Integer);
  var
    I,J: Integer;
    D,T: TColorRec;
  begin
    I := L;
    repeat
      L := I;
      J := R;
      D := Colors[(L + R) shr 1];
      repeat
        while Compare(Colors[I], D) < 0 do Inc(I);
        while Compare(Colors[J], D) > 0 do Dec(J);
        if I > J then Break;
        T    := Colors[I];
        Colors[I] := Colors[J];
        Colors[J] := T;
        Inc(I);
        Dec(J);
      until I > J;
      if L < J then Sort(Colors, L, J);
    until I >= R;
  end;

var
  D: TBitmap;
  X,Y,I,J,C: Integer;
  L: PWordArray;
  V: Word;
begin
  Result := nil;
  D := Image2.Picture.Bitmap;
  D.PixelFormat := pf16Bit;
  I := -1;
  C := 0;
  for Y := 0 to D.Height -1 do
  begin
    L := D.ScanLine[Y];
    for X := 0 to D.Width -1 do
    begin
      V := L[X];
      if (I < 0) or (Result[I].Value <> V) then
      begin
        I := -1;
        for J := 0 to C -1 do
          if Result[J].Value = V then
          begin
            I := J;
            Break;
          end;
        if I < 0 then
        begin
          I := C;
          if C mod 256 = 0 then SetLength(Result, C + 256);
          Result[C].Value := V;
          Inc(C);
        end;
      end;
      Inc(Result[I].Count);
    end;
  end;
  SetLength(Result, C);
  Sort(Result, 0, C -1);
end;


procedure TForm1.Compress;
var
  Data,DataBits: Cardinal;
{
  Offsets: TOffsetTable;
  OffsetCount: Integer;

type
  TOffsetRec = record
                 Value: Integer;
                 Count: Integer;
               end;
  TOffsetTable = array of TOffsetRec;

  procedure Sort(Offsets: TOffsetTable; L,R: Integer);
  var
    I,J: Integer;
    D,T: TOffsetRec;
  begin
    I := L;
    repeat
      L := I;
      J := R;
      D := Offsets[(L + R) shr 1];
      repeat
        while Offsets[I].Count > D.Count do Inc(I);
        while Offsets[J].Count < D.Count do Dec(J);
        if I > J then Break;
        T    := Offsets[I];
        Offsets[I] := Offsets[J];
        Offsets[J] := T;
        Inc(I);
        Dec(J);
      until I > J;
      if L < J then Sort(Offsets, L, J);
    until I >= R;
  end;

  procedure SortX(Offsets: TOffsetTable; L,R: Integer);
  var
    I,J: Integer;
    D,T: TOffsetRec;
  begin
    I := L;
    repeat
      L := I;
      J := R;
      D := Offsets[(L + R) shr 1];
      repeat
        while Abs(Offsets[I].Value) < Abs(D.Value) do Inc(I);
        while Abs(Offsets[J].Value) > Abs(D.Value) do Dec(J);
        if I > J then Break;
        T    := Offsets[I];
        Offsets[I] := Offsets[J];
        Offsets[J] := T;
        Inc(I);
        Dec(J);
      until I > J;
      if L < J then SortX(Offsets, L, J);
    until I >= R;
  end;

  procedure AddOffset(Value: Integer);
  var
    I: Integer;
  begin
    for I := 0 to High(Offsets) do
      if Offsets[I].Value = Value then
      begin
        Inc(Offsets[I].Count);
        Exit;
      end;
    if OffsetCount mod 256 = 0 then SetLength(Offsets, OffsetCount + 256);
    Offsets[OffsetCount].Value := Value;
    Offsets[OffsetCount].Count := 1;
    Inc(OffsetCount);
  end;
}
  procedure WriteByte(Value: Byte);
  begin
    FStream.Write(Value, SizeOf(Value));
  end;

  procedure WriteWord(Value: Word);
  begin
//    Value := Swap(Value);
    FStream.Write(Value, SizeOf(Value));
  end;

  procedure AddBits(Value,Bits: Cardinal);
  begin
    Data := Data or (Value shl DataBits);
    Inc(DataBits, Bits);
    while DataBits >= 8 do
    begin
      Dec(DataBits, 8);
      WriteByte(Data);
      Data := Data shr 8;
    end;
  end;

// Monochrome Bitmaps use RLE Encoding, 4 Bits Count + 4 Pixelbits
var
  NibbleBits,NibbleCount: Integer;
  Nibble,LastNibble: Byte;

  procedure SaveNibble;
  begin
    while NibbleCount >= 16 do
    begin
      WriteByte($F0 or LastNibble);
      Dec(NibbleCount, 16);
    end;
    if NibbleCount > 0 then
      WriteByte((NibbleCount -1) shl 4 or LastNibble);
    NibbleCount := 0;
  end;

  procedure AddNibble(Bit: Boolean);
  begin
    Inc(Nibble, Nibble);
    if Bit then Inc(Nibble);
    Inc(NibbleBits);
    if NibbleBits = 4 then
    begin
      if Nibble <> LastNibble then
      begin
        if LastNibble <> $FF then
          SaveNibble;
        LastNibble := Nibble;
      end;
      Inc(NibbleCount);
      NibbleBits := 0;
      Nibble := 0;
    end;
  end;

var
  Colors: TColorTable;
  D: TBitmap;
  L: PWordArray;
  X,Y,I: Integer;
  P: Word;
  BPP: Byte;
label
  Start;
begin
  M.Clear;
  Colors := ColorTable;
  D := Image2.Picture.Bitmap;
  P := PWordArray(D.ScanLine[0])[0];
  BPP := 0;
  while (1 shl BPP) < Length(Colors) do Inc(BPP);
Start:
  Data := 0;
  DataBits := 0;
  FreeAndNil(FStream);
  FStream := TMemoryStream.Create;
  WriteByte(D.Width);
  WriteByte(D.Height);
  WriteByte(BPP);
  if BPP = 0 then
  begin // FillRect, only one Color
    WriteWord(P);
  end else
    if BPP = 1 then
    begin // Monochrome, RLE encoded
      WriteWord(P);                                             // colortable
      if Colors[0].Value <> P then WriteWord(Colors[0].Value)   // only two colors
        else WriteWord(Colors[1].Value);
      NibbleBits := 0;
      NibbleCount := 0;
      LastNibble := $FF;
      Nibble := 0;
      for Y := 0 to D.Height -1 do
      begin
        L := D.ScanLine[Y];
        for X := 0 to D.Width -1 do
          AddNibble(L[X] <> P);
      end;
      SaveNibble;
    end else
      if BPP < 16 then
      begin // color
        WriteWord(P);
        WriteWord(Length(Colors) * 2);
        for I := 0 to High(Colors) do
          WriteWord(Colors[I].Value);
        for Y := 0 to D.Height -1 do
        begin
          L := D.ScanLine[Y];
          for X := 0 to D.Width -1 do
            if P <> L[X] then
            begin
              P := L[X];
              for I := 0 to High(Colors) do
                if Colors[I].Value = P then
                begin
                  AddBits(1, 1);
                  AddBits(I, BPP);
                  Break;
                end;
            end else
              if (Y > 0) or (X > 0) then
                AddBits(0, 1);
        end;
      end else
      begin // 16 Bit full color, uncompressed without colortable
        for Y := 0 to D.Height -1 do
        begin
          L := D.ScanLine[Y];
          for X := 0 to D.Width -1 do
            WriteWord(L[X]);
        end;
      end;
  if DataBits > 0 then AddBits(0, DataBits + 8 - Databits mod 8);
  if (BPP < 16) and (FStream.Size >= D.Width * D.Height * 2) then
  begin
    BPP := 16;
    goto Start;
  end;
  M.Lines.Add(Format('Pixelcount    : %7d', [D.Width * D.Height]));
  M.Lines.Add(Format('Colorcount    : %7d', [Length(Colors)]));
  M.Lines.Add(Format('RAW Size      : %7d byte', [D.Width * D.Height * 2]));
  M.Lines.Add(Format('Compressed    : %7d byte', [FStream.Size]));
  M.Lines.Add(Format('Ratio         : %7s %%', [FormatFloat('#0', D.Width * D.Height * 2 / FStream.Size * 100)]));
  M.Lines.Add(Format('BPP           : %7d', [BPP]));
end;

procedure TForm1.Convert;
type
  TRGB = packed record
    R,G,B: Integer;
  end;

  function GetRGB(C: Word): TRGB;
  begin
    Result.R := C shr 11;
    Result.G := C shr 5 and $3F;
    Result.B := C and $1F;
  end;

var
  S,D: TBitmap;
  X,Y,I,J,K,R,T,U,V,W: Integer;
  C: TColorTable;
  AL,AH,AM,B: TRGB;
  L: PWordArray;
  F: array of Integer;
  P: Word;
begin
  C := nil;
  F := nil;
  if Image1.Picture.Graphic = nil then Exit;
  S := Image1.Picture.Bitmap;
  D := TBitmap.Create;
  try
    S.PixelFormat := pf24Bit;
    D.PixelFormat := pf16Bit;
    D.Width := S.Width;
    D.Height := S.Height;
    for Y := 0 to D.Height -1 do
      for X := 0 to D.Width -1 do
        D.Canvas.Pixels[X, Y] := S.Canvas.Pixels[X, Y];

    U := StrToIntDef(Edit1.Text, 1) +1;
    for V := 1 to U -1 do
    begin
      Image2.Picture.Bitmap.Assign(D);
      C := ColorTable;
      SetLength(F, Length(C));
      for I := 0 to High(C) do
        C[I].NewValue := C[I].Value;
      for I := 0 to High(C) -1 do
        if C[I].Value = C[I].NewValue then
        begin
          AL := GetRGB(C[I].Value);
          AH := AL;
          AM := AL;
          W := (V +1) div 2;
          Inc(AH.R, W);
          Inc(AH.G, W);
          Inc(AH.B, W);
          W := V div 4;
          Inc(AM.R, W);
          Inc(AM.G, W);
          Inc(AM.B, W);
          F[0] := I;
          K := 1;
          for J := 0 to High(C) do
            if (C[J].Value = C[J].NewValue) and (C[J].Count <= C[I].Count) then
            begin
              B := GetRGB(C[J].Value);
              if ((B.R >= AL.R) and (B.R <= AH.R) and (B.G >= AL.G) and (B.G <= AM.G) and (B.B >= AL.B) and (B.B <= AM.B)) or
                 ((B.R >= AL.R) and (B.R <= AM.R) and (B.G >= AL.G) and (B.G <= AH.G) and (B.B >= AL.B) and (B.B <= AM.B)) or
                 ((B.R >= AL.R) and (B.R <= AM.R) and (B.G >= AL.G) and (B.G <= AM.G) and (B.B >= AL.B) and (B.B <= AH.B)) then
              begin
                F[K] := J;
                Inc(K);
              end;
            end;
          R := 0;
          AL.R := 0;
          AL.G := 0;
          AL.B := 0;
          for J := 0 to K -1 do
          begin
            T := C[F[J]].Count;
            Inc(R, T);
            B := GetRGB(C[F[J]].Value);
            Inc(AL.R, B.R * T);
            Inc(AL.G, B.G * T);
            Inc(AL.B, B.B * T);
          end;
          AL.R := Round(AL.R / R) and $1F;
          AL.G := Round(AL.G / R) and $3F;
          AL.B := Round(AL.B / R) and $1F;
          P := AL.R shl 11 or AL.G shl 5 or AL.B;
          for J := 0 to K -1 do
            C[F[J]].NewValue := P;
        end;

      for Y := 0 to D.Height -1 do
      begin
        L := D.ScanLine[Y];
        for X := 0 to D.Width -1 do
        begin
          for I := 0 to High(C) do
            if L[X] = C[I].Value then
            begin
              L[X] := C[I].NewValue;
              Break;
            end;
        end;
      end;
    end;
    Image2.Picture.Bitmap.Assign(D);
  finally
    D.Free;
  end;
  Label1.Caption := Format('24Bit, Width: %d, Height: %d', [Image1.Picture.Width, Image1.Picture.Height]);
  Label2.Caption := Format('16Bit, Width: %d, Height: %d', [Image2.Picture.Width, Image2.Picture.Height]);
  Compress;
end;

procedure TForm1.Button1Click(Sender: TObject);
var
  P: TPicture;
  B: TBitmap;
  R: TRect;
begin
  if OPD.Execute then
  begin
    B := TBitmap.Create;
    try
      P := TPicture.Create;
      try
        P.LoadFromFile(OPD.FileName);
        R := Bounds(0, 0, P.Width, P.Height);
        if (R.Right > R.Bottom) and ((R.Right > 176) or (R.Bottom > 132)) then
        begin
          R.Bottom := Trunc(R.Bottom / R.Right * 176);
          R.Right := 176;
          if R.Bottom > 132 then
          begin
            R.Right := Trunc(R.Right / R.Bottom * 132);
            R.Bottom := 132;
          end;
        end else
          if (R.Bottom >= R.Right) and ((R.Bottom > 176) or (R.Right > 132)) then
          begin
            R.Right := Trunc(R.Right / R.Bottom * 176);
            R.Bottom := 176;
            if R.Right > 132 then
            begin
              R.Bottom := Trunc(R.Bottom / R.Right * 132);
              R.Right := 132;
            end;
          end;
        B.Width := R.Right;
        B.Height := R.Bottom;
        B.Canvas.StretchDraw(R, P.Graphic);
      finally
        P.Free;
      end;
      Image1.Picture.Bitmap.Assign(B);
      Button3.Enabled := Image1.Picture.Graphic <> nil;
      Button4.Enabled := Button3.Enabled;
      Button2.Enabled := Button3.Enabled;
      CB.Enabled := Button3.Enabled;
      Edit1.Enabled := Button3.Enabled;
      UpDown1.Enabled := Button3.Enabled;
      Label3.Enabled := Button3.Enabled;
    finally
      B.Free;
    end;
    Convert;
  end;
end;

procedure TForm1.Button3Click(Sender: TObject);
var
  S,D: TBitmap;
  X,Y: Integer;
begin
  S := Image1.Picture.Bitmap;
  D := TBitmap.Create;
  try
    D.Width := S.Height;
    D.Height := S.Width;
    for Y := 0 to S.Height -1 do
      for X := 0 to S.Width -1 do
        D.Canvas.Pixels[S.Height -1 - Y, X] := S.Canvas.Pixels[X, Y];
    Image1.Picture.Bitmap.Assign(D);
  finally
    D.Free;
  end;
  Convert;
end;

procedure TForm1.Button4Click(Sender: TObject);
var
  S,D: TBitmap;
  X,Y: Integer;
begin
  S := Image1.Picture.Bitmap;
  D := TBitmap.Create;
  try
    D.Width := S.Width;
    D.Height := S.Height;
    for Y := 0 to S.Height -1 do
      for X := 0 to S.Width -1 do
        D.Canvas.Pixels[S.Width -1 - X, Y] := S.Canvas.Pixels[X, Y];
    Image1.Picture.Bitmap.Assign(D);
  finally
    D.Free;
  end;
  Convert;
end;

procedure TForm1.Button2Click(Sender: TObject);
const
  CR     = #13#10;

  sData  = '/*'                                                         + CR +
           '   created with S65 Image Converter written by H. Reddmann' + CR +
           '   HaReddmann at t-online dot de'                           + CR + CR +
           '   Filename         : %0:s'                                 + CR +
           '   Bitmaptype       : %2:s'                                 + CR +
           '   Width            : %3:d'                                 + CR +
           '   Height           : %4:d'                                 + CR +
           '   Number of Colors : %5:d'                                 + CR +
           '   Bits Per Pixel   : %6:d'                                 + CR +
           '   Colortable       : %7:s'                                 + CR +
           '   Size             : %8:d bytes'                           + CR +
           '   Imagesize        : %9:d bytes'                           + CR +
           '*/'                                                         + CR + CR +
           '#ifndef %1:s_H'                                             + CR +
           '#define %1:s_H'                                             + CR + CR +
           '#include <inttypes.h>'                                      + CR +
           '#include <avr/pgmspace.h>'                                  + CR + CR +
           '#define %1:s_WIDTH  %3:d'                                   + CR +
           '#define %1:s_HEIGHT %4:d'                                   + CR + CR +
           'uint8_t %1:s_bmp[] PROGMEM = {'                             +
           '%10:s};'                                                    + CR + CR +
           '#endif'                                                     + CR;

   sNonePresent: array[Boolean] of String = ('none', 'present');

var
  Width,Height,BPP: Byte;

  function GetRAW: String;
  var
    D: TBitmap;
    X,Y,C: Integer;
    L: PWordArray;
  begin
    D := Image2.Picture.Bitmap;
    Width := D.Width;
    Height := D.Height;
    BPP := 16;
    Result := '';
    C := 0;
    for Y := 0 to D.Height -1 do
    begin
      L := D.ScanLine[Y];
      for X := 0 to D.Width -1 do
      begin
        if C mod 8 = 0 then Result := Result + CR + '  ';
        Inc(C);
        Result := Result + Format('0x%0.2x, 0x%0.2x, ', [L[X] shr 8, L[X] and $FF]);
      end;
    end;
  end;

  function GetStreamData(Count: Integer): String;
  var
    C: Integer;
    D: Byte;
  begin
    Result := '';
    C := 0;
    while (Count > 0) and (FStream.Position < FStream.Size) do
    begin
      Dec(Count);
      if C mod 16 = 0 then Result := Result + CR + '  ';
      Inc(C);
      FStream.Read(D, SizeOf(D));
      Result := Result + Format('0x%0.2x, ', [D]);
    end;
  end;

var
  Data: String;
  Colors: TColorTable;
  BitmapType: String;
  Size,I: Integer;
  ColorSize: Word;
begin
  Colors := ColorTable;
  if SD.Execute then
  begin
    if CB.Checked then
    begin
      BitmapType := 'RAW';
      Data := GetRaw;
      Size := Width * Height * 2;
    end else
    begin
      Size := FStream.Size;
      FStream.SaveToFile(ChangeFileExt(SD.FileName, '.bin'));
      FStream.Position := 0;
      FStream.Read(Width, SizeOf(Width));
      FStream.Read(Height, SizeOf(Height));
      FStream.Read(BPP, SizeOf(BPP));
      if BPP = 0 then
      begin
        BitmapType := 'one Color FillRect';
        FStream.Position := 0;
        Data := GetStreamData(3) + CR + GetStreamData(MaxInt);
      end else
        if BPP = 1 then
        begin
          BitmapType := 'Monochrome, compressed';
          FStream.Position := 0;
          Data := GetStreamData(7) + CR + GetStreamData(MaxInt);
        end else
          if BPP = 16 then
          begin
            BitmapType := 'Fullcolor, uncompressed';
            FStream.Position := 0;
            Data := GetStreamData(3) + CR + GetStreamData(MaxInt);
          end else
          begin
            BitmapType := 'Color, compressed';
            FStream.Read(ColorSize, SizeOf(ColorSize));
            FStream.Read(ColorSize, SizeOf(ColorSize));
   //         ColorSize := Swap(ColorSize);
            FStream.Position := 0;
            Data := GetStreamData(7) + CR + GetStreamData(ColorSize) + CR + GetStreamData(MaxInt);
          end;
    end;
    I := Length(Data);
    while Data[I] in [',', ' '] do Dec(I);
    SetLength(Data, I);
    Data := Format(sData, [ExtractFileName(SD.FileName),
                           ChangeFileExt(ExtractFileName(SD.FileName), ''),
                           BitmapType,
                           Width,
                           Height,
                           Length(Colors),
                           BPP,
                           sNonePresent[(BPP > 1) and (BPP < 16)],
                           Size,
                           Width * Height * 2,
                           Data]);

    with TFileStream.Create(SD.FileName, fmCreate) do
    try
      Write(PChar(Data)^, Length(Data));
    finally
      Free;
    end;
  end;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FreeAndNil(FStream);
end;

procedure TForm1.CBColorClick(Sender: TObject);
begin
  Convert;
end;

end.
